/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2011 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.base;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.conqat.ide.commons.ui.extension.ExtensionPointUtils;
import org.conqat.ide.commons.ui.logging.LoggingUtils;
import org.conqat.lib.commons.reflect.ReflectionUtils;
import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.emf.ecore.EObject;
import org.fortiss.tooling.kernel.ToolingKernelActivator;
import org.fortiss.tooling.kernel.interfaces.IEObjectAware2;
import org.osgi.framework.Bundle;

/**
 * Service base implementation, which supports handler registration with two
 * related {@link EObject} classes.
 * 
 * @author hoelzlf
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: F6B36995BE099A30EA854295563C109E
 */
public abstract class EObjectAware2ServiceBase<T extends IEObjectAware2<? extends EObject, ? extends EObject>> {

	/** Stores the source class to target class to handler map. */
	protected final Map<Class<?>, Map<Class<?>, List<T>>> handlersBySource = new HashMap<Class<?>, Map<Class<?>, List<T>>>();

	/** Stores the target class to source class to handler map. */
	protected final Map<Class<?>, Map<Class<?>, List<T>>> handlersByTarget = new HashMap<Class<?>, Map<Class<?>, List<T>>>();

	/** Constructor. */
	public EObjectAware2ServiceBase() {
		setupHandlerMap();
	}

	/** Returns the extension point name. */
	protected abstract String getExtensionPointName();

	/** Returns the configuration element name. */
	protected abstract String getConfigurationElementName();

	/** Returns the handler class attribute name. */
	protected abstract String getHandlerClassAttribute();

	/**
	 * Returns the source attribute name. Sub-classes may override, but should
	 * first consider to use the default value <i>source</i> in the extension
	 * point definition.
	 */
	protected String getSourceAttribute() {
		return "source";
	}

	/**
	 * Returns the target attribute name. Sub-classes may override, but should
	 * first consider to use the default value <i>target</i> in the extension
	 * point definition.
	 */
	protected String getTargetAttribute() {
		return "target";
	}

	/**
	 * Returns the attribute name for the model element class. Sub-classes may
	 * override, but should first consider to use the default value
	 * <i>modelElementClass</i> in the extension point definition.
	 */
	protected String getModelElementClassAttribute() {
		return "modelElementClass";
	}

	/**
	 * Returns whether the service requires a unique handler per model element
	 * class. By default more than one handler can be registered per model
	 * element class.
	 */
	protected boolean useUniqueHandler() {
		return false;
	}

	/** Returns the list of registered handlers for the given class. */
	public List<T> getRegisteredHandlers(Class<?> sourceModelElementClass,
			Class<?> targetModelElementClass) {
		Map<Class<?>, List<T>> innerMap = ReflectionUtils
				.performNearestClassLookup(sourceModelElementClass,
						handlersBySource);
		if (innerMap == null) {
			return null;
		}
		return ReflectionUtils.performNearestClassLookup(
				targetModelElementClass, innerMap);
	}

	/** Initializes the handler map from plugin extensions. */
	private void setupHandlerMap() {
		for (IConfigurationElement ce : ExtensionPointUtils
				.getConfigurationElements(getExtensionPointName(),
						getConfigurationElementName())) {
			try {
				createHandler(ce);
			} catch (Exception e) {
				LoggingUtils.error(ToolingKernelActivator.getDefault(),
						"Error reading connector extension!", e);
			}
		}
	}

	/** Creates the handler from the given configuration element. */
	@SuppressWarnings("unchecked")
	private void createHandler(IConfigurationElement ce) throws Exception {
		Bundle bundle = ExtensionPointUtils.getBundle(ce);

		Class<?> handlerClass = ExtensionPointUtils.loadClass(
				ce.getAttribute(getHandlerClassAttribute()), bundle);
		T handler = (T) handlerClass.getConstructor().newInstance();

		List<Class<?>> sources = new ArrayList<Class<?>>();
		List<Class<?>> targets = new ArrayList<Class<?>>();

		for (final IConfigurationElement child : ce.getChildren()) {
			if (child.getName().equals(getSourceAttribute())) {
				sources.add(ExtensionPointUtils.loadClass(
						child.getAttribute(getModelElementClassAttribute()),
						bundle));
			} else if (child.getName().equals(getTargetAttribute())) {
				targets.add(ExtensionPointUtils.loadClass(
						child.getAttribute(getModelElementClassAttribute()),
						bundle));
			}
		}

		for (final Class<?> source : sources) {
			for (final Class<?> destination : targets) {
				addHandler(source, destination, handler);
			}
		}
	}

	/** Adds the given handler to the map. */
	private void addHandler(Class<?> sourceElementClass,
			Class<?> targetElementClass, T handler) {

		insertStorage(handlersBySource, sourceElementClass, targetElementClass,
				handler);
		insertStorage(handlersByTarget, targetElementClass, sourceElementClass,
				handler);
	}

	/** Performs storage into the maps. */
	private void insertStorage(Map<Class<?>, Map<Class<?>, List<T>>> map,
			Class<?> key1, Class<?> key2, T handler) {

		Map<Class<?>, List<T>> innerMap = map.get(key1);
		if (innerMap == null) {
			innerMap = new HashMap<Class<?>, List<T>>();
			map.put(key1, innerMap);
		}

		List<T> list = innerMap.get(key2);
		if (list == null) {
			list = new ArrayList<T>();
			innerMap.put(key2, list);
		}

		if (useUniqueHandler() && !list.isEmpty()) {
			LoggingUtils.log(
					ToolingKernelActivator.getDefault(),
					"Encountered more than one handler registered with "
							+ getExtensionPointName() + "("
							+ handler.getClass() + "," + list.get(0).getClass()
							+ ")", IStatus.WARNING);
		}
		list.add(handler);
	}
}
